#!/usr/bin/python3
# A mini-app for accepting Zoom webhooks

import configparser
import json
import psycopg2
import psycopg2.extras
import traceback
import dateutil.parser
# import dateutil.tz

### Configuration ###

config = configparser.ConfigParser()
config.read('zoom.ini')

### Database connection ###

db_connection = None
db = None

def db_connect():
    global db_connection, db
    db_connection = psycopg2.connect(
                host='localhost',
                user=config['db']['user'],
                password=config['db']['passwd'],
                dbname=config['db']['name'],
            )
    db = db_connection.cursor(cursor_factory=psycopg2.extras.NamedTupleCursor)

def db_query(query, args=()):
    if db is None:
        db_connect()
    try:
        db.execute(query, args)
    except psycopg2.DatabaseError:
        # Reconnect if the connection died (timeout, server restart etc.)
        db_connect()
        db.execute(query, args)

### Utilities ###

# FIXME: Move to shared code?

def parse_time(iso_time, tz_name):
    utc = dateutil.parser.isoparse(iso_time)
    # tz = dateutil.tz.gettz(tz_name)
    # return utc.astimezone(tz)
    return utc

### Application ###

class HookApp:

    def __init__(self, env, start_response):
        self.env = env
        self.wsgi_start = start_response

    def log(self, msg):
        print(msg, file=self.env['wsgi.errors'], flush=True)

    def http_error(self, code, msg, extra_headers=[]):
        self.wsgi_start("{} {}".format(code, msg), extra_headers + [
                ('Content-Type', 'text/plain')
            ])
        return ["{} {}".format(code, msg)]

    def create_regular_meeting(self, uid, meeting):
        meeting_id = meeting["id"]
        self.log(f"Meeting {meeting_id}: Planning")

        db_query("""
                INSERT INTO zoom_meetings
                (meeting_id, uuid, host_id, topic, type, start_time, duration)
                VALUES (%s, %s, %s, %s, %s, %s, %s)
            """, (
                meeting_id,
                meeting['uuid'],
                uid,
                meeting['topic'],
                meeting['type'],
                parse_time(meeting['start_time'], meeting['timezone']),
                meeting['duration'],
            ))

        db_connection.commit()


    def create_recurring_meeting(self, uid, meeting):
        meeting_id = meeting["id"]

        for occ in meeting["occurrences"]:
            occ_id = occ["occurrence_id"]
            self.log(f"Meeting {meeting_id}: Planning occurrence {occ_id}")

            db_query("""
                    INSERT INTO zoom_meetings
                    (meeting_id, uuid, occurrence_id, host_id, topic, type, start_time, duration)
                    VALUES (%s, %s, %s, %s, %s, %s, %s, %s)
                """, (
                    meeting_id,
                    meeting['uuid'],
                    occ_id,
                    uid,
                    meeting['topic'],
                    meeting['type'],
                    parse_time(occ['start_time'], meeting['timezone']),
                    occ['duration'],
                ))

        db_connection.commit()


    def create_meeting(self, js):
        payload = js["payload"]
        meeting = payload["object"]
        meeting_id = meeting["id"]

        host_user_id = meeting["host_id"]
        db_query("SELECT * FROM zoom_users WHERE user_id=%s", (host_user_id,))
        user = db.fetchone()
        if user is None:
            self.log(f"Meeting {meeting_id}: Host {host_user_id} not found in zoom_users")
            return

        type = meeting["type"]
        if type == 2:
            self.create_regular_meeting(user.id, meeting)
        elif type == 8:
            self.create_recurring_meeting(user.id, meeting)
        else:
            self.log(f"Meeting {meeting_id}: Unknown type {type}")
            return


    def delete_regular_meeting(self, meeting):
        meeting_id = meeting["id"]

        self.log(f"Meeting {meeting_id}: Deleting")
        db_query("""
                DELETE FROM zoom_meetings
                WHERE meeting_id=%s
            """, (
                meeting_id,
            ))

        db_connection.commit()


    def delete_recurring_meeting(self, meeting):
        meeting_id = meeting["id"]

        if "occurrences" not in meeting:
            return self.delete_regular_meeting(meeting)

        for occ in meeting["occurrences"]:
            occ_id = occ["occurrence_id"]
            self.log(f"Meeting {meeting_id}: Deleting occurrence {occ_id}")
            db_query("""
                    DELETE FROM zoom_meetings
                    WHERE meeting_id=%s AND occurrence_id=%s
                """, (
                    meeting_id,
                    occ_id,
                ))

        db_connection.commit()


    def delete_meeting(self, js):
        payload = js["payload"]
        meeting = payload["object"]
        meeting_id = meeting["id"]

        type = meeting["type"]
        if type == 2:
            self.delete_regular_meeting(meeting)
        elif type == 8:
            self.delete_recurring_meeting(meeting)
        else:
            self.log(f"Meeting {meeting_id}: Unknown type {type}")
            return


    def run(self):
        method = self.env['REQUEST_METHOD']
        if method != 'POST':
            return self.http_error(405, 'Method not allowed', [('Allow', 'POST')])

        if self.env.get('HTTP_AUTHORIZATION', '') != config['hooks']['verification_token']:
            self.log('Verification token does not match!')
            return self.http_error(401, 'Authorization failed')

        body = self.env['wsgi.input'].read()
        js = json.loads(body)
        self.log(js)

        event = js["event"]
        if event == "meeting.created":
            self.create_meeting(js)
        elif event == "meeting.deleted":
            self.delete_meeting(js)
        else:
            self.log(f"Unknown event: {event}")

        self.wsgi_start("204 No Content", [])
        return b""

def application(env, start_response):
    app = HookApp(env, start_response)
    try:
        return app.run()
    except Exception as exc:
        app.log(traceback.print_exception(etype=None, value=exc, tb=exc.__traceback__))
        return app.http_error(500, "Internal server error")

# FIXME: Number of workers and access rights; reloading
